##############################################################################
## swigprocess.py -- Script to generate swig based interface for python and R
## Revision: $Id$
##
## Copyright (C) Kshitij Kulshreshtha
##
## This file is part of ADOL-C. This software is provided as open source.
## Any use, reproduction, or distribution of the software constitutes 
## recipient's acceptance of the terms of the accompanying license file.
## 
##############################################################################

import re
import os.path
import sys
import argparse
import glob
import shlex
import shutil
import subprocess
import atexit

noerrors = True

@atexit.register
def call_make_clean():
    if noerrors:
        pass
    elif len(glob.glob('[Mm]akefile')) > 0:
        subprocess.call(['make','clean'])

def readFile(name):
    alllines = ''
    inputfile= open(name, 'r')
    for line in inputfile:
        alllines+=line
    inputfile.close()
    return alllines

def writeOutput(outstr, outfile):
    outputfile= open(outfile, 'w')
    outputfile.write(outstr)
    outputfile.close()


def appendFile(outfile, outstr):
    outputfile = open(outfile,'a')
    outputfile.write(outstr)
    outputfile.write('\n')
    outputfile.close()

def comment_all_includes(lines):
    s = r'(#\s*include\s*<\S*>)'
    p = re.compile(s,re.M|re.S)
    newlines = p.sub(r'//\1',lines)
    return newlines

def uncomment_local_includes(lines):
    s = r'//(#\s*include\s*<adolc/\S*>)'
    p = re.compile(s,re.M|re.S)
    newlines = p.sub(r'\1',lines)
    return newlines

def invoke_cpp(infile,outfile):
    s = os.environ['CXX'] + ' -std=c++11 -E -C -P -o ' + outfile + ' -Iinclude -nostdinc -DSWIGPRE ' + infile
    print('invoking:', s)
    cmd = shlex.split(s)
    try:
        warn = subprocess.check_output(cmd,universal_newlines=True)
    except subprocess.CalledProcessError as e:
        print(e.output)
        print("error in cmd = ", e.cmd)
        exit()
    if len(warn) > 0:
        print(warn)

def reinstate_nonlocal_include(lines):
    s = r'//(#\s*include\s*<\S*>)'
    p = re.compile(s,re.M|re.S)
    newlines = p.sub(r'\1',lines)
    return newlines    

def cleanup(intfile):
    shutil.rmtree('include')
    os.remove(intfile)


def invoke_swig_compile(lang,infile,outfile,modname):
    try:
        os.mkdir(lang)
    except:
        shutil.rmtree(lang)
        os.mkdir(lang)

    if lang == 'R':
        s = 'swig -r -c++ -o ' + outfile + ' ' + infile
        print('invoking:', s)
        cmd = shlex.split(s)
        warn = ''
        try:
            warn += subprocess.check_output(cmd,stderr=subprocess.STDOUT,universal_newlines=True)
        except subprocess.CalledProcessError as e:
            print(e.output)
            print("error in cmd = ", e.cmd)
            exit()
        s = 'R CMD SHLIB -o ' + modname + '.so ' + outfile + ' -L../.libs -ladolc'
        if sys.platform.startswith('linux'):
            s += ' -Wl,-no-undefined'
        evars = os.environ 
        evars['PKG_CPPFLAGS'] = "-I../include -std=c++11" 
        print('invoking:', s)
        cmd = shlex.split(s)
        try:
            warn += subprocess.check_output(cmd,env=evars,stderr=subprocess.STDOUT,universal_newlines=True)
        except subprocess.CalledProcessError as e:
            print(e.output)
            print("error in cmd = ", e.cmd)
            exit()            
        shutil.move(modname + '.so', lang)
        shutil.move(modname + '.R', lang)
        shutil.move(outfile,lang)
        writeOutput(warn,'warnings-r.txt')
    elif lang == 'octave':
        s = 'swig -octave -c++ -o ' + outfile + ' ' + infile
        print('invoking:', s)
        cmd = shlex.split(s)
        warn = ''
        try:
            warn += subprocess.check_output(cmd,stderr=subprocess.STDOUT,universal_newlines=True)
        except subprocess.CalledProcessError as e:
            print(e.output)
            print("error in cmd = ", e.cmd)
            exit()
        s = 'mkoctfile -o ' + modname + '.oct ' + ' -I../include -std=c++11'  + ' -L../.libs -ladolc ' + outfile
        if sys.platform.startswith('linux'):
            s += ' -Wl,-no-undefined '
        print('invoking:', s)
        cmd = shlex.split(s)
        try:
            warn += subprocess.check_output(cmd,stderr=subprocess.STDOUT,universal_newlines=True)
        except subprocess.CalledProcessError as e:
            print(e.output)
            print("error in cmd = ", e.cmd)
            exit()            
        shutil.move(modname + '.oct', lang)
        shutil.move(outfile,lang)
        writeOutput(warn,'warnings-octave.txt')
    elif lang == 'python':
        python_cflags = subprocess.check_output(['python-config','--cflags'],universal_newlines=True)
        python_ldflags = subprocess.check_output(['python-config','--ldflags'],universal_newlines=True)
        from numpy.distutils import misc_util as npy_dist
        incp = npy_dist.get_numpy_include_dirs()
        npy_cflags = ''
        for p in incp:
            npy_cflags += ' -I' + p 

        s = 'swig -python -c++ -dirvtable -o ' + outfile + ' ' + infile
        print('invoking:', s)
        cmd = shlex.split(s)
        warn = ''
        try:
            warn += subprocess.check_output(cmd,stderr=subprocess.STDOUT,universal_newlines=True)
        except subprocess.CalledProcessError as e:
            print(e.output)
            print("error in cmd = ", e.cmd)
        p = re.compile(r'(.*)\.cxx',re.M|re.S)
        outhead = p.sub(r'\1.h',outfile)
        s = os.environ['CXX'] + ' -I../include -std=c++11 -fPIC -Wall -shared -o _' + modname + '.so ' + python_cflags.rstrip() + npy_cflags + ' ' + outfile + ' -L../.libs -ladolc ' + python_ldflags.rstrip() 
        if sys.platform.startswith('linux'):
            s += ' -Wl,-no-undefined'
        print('invoking:', s)
        cmd = shlex.split(s)
        try:
            warn += subprocess.check_output(cmd,stderr=subprocess.STDOUT,universal_newlines=True)
        except subprocess.CalledProcessError as e:
            print(e.output)
            print("error in cmd = ", e.cmd)
            exit()            
        shutil.move('_' + modname + '.so', lang)
        shutil.move(modname + '.py', lang)
        shutil.move(outfile,lang)
        shutil.move(outhead,lang)
        writeOutput(warn,'warnings-python.txt')

def finalClean(headfile,outfiles):
    if os.path.isfile(headfile):
        os.remove(headfile)
    for f in outfiles:
        if os.path.isfile(f):
            os.remove(f)
    for f in glob.glob('*.o'):
        os.remove(f)

def prepare_flat_header():
    sys.path = [ os.getcwd() ] + sys.path
    p = os.getcwd() + '/../include/adolc'
    for (dp, dn, fn) in os.walk(p):
        ndp = re.sub(r'\.\./',r'',dp)
        for f in iter(fn):
            lines = readFile(dp + "/" + f)
            lines = comment_all_includes(lines)
            lines = uncomment_local_includes(lines)
            try:
                os.makedirs(ndp)
            except:
                pass
            writeOutput(lines, ndp + "/" + f)
    
    invoke_cpp('adolc_all_in.hpp', 'adolc_all_pre.hpp')
    lines = readFile('adolc_all_pre.hpp')
    lines = reinstate_nonlocal_include(lines)
    writeOutput(lines,'adolc_all.hpp')
    cleanup('adolc_all_pre.hpp')


def main(args):
    global noerrors
    noerrors = False
    prepare_flat_header()
    if args.py or args.all:
        invoke_swig_compile('python','adolc-python.i','adolc_python_wrap.cxx','adolc')
    if args.r or args.all:
        invoke_swig_compile('R','adolc-r.i','adolc_r_wrap.cpp','adolc')
    if args.oc or args.all:
        invoke_swig_compile('octave','adolc-octave.i','adolc_octave_wrap.cpp','adolc')
    finalClean('adolc_all.hpp',['adolc_python_wrap.cxx','adolc_python_wrap.h','adolc_r_wrap.cpp'])
    noerrors = True

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Compile Swig Interfaces')
    parser.add_argument('--py', '--python', action='store_true', 
                        help='compile python interface')
    parser.add_argument('--r', action='store_true',
                        help='compile R interface')
    parser.add_argument('--oc', '--octave', action='store_true',
                        help='compile Octave interface')
    parser.add_argument('--all', action='store_true', default=True,
                        help='compile all interfaces (python, R, octave) [default]')
    args = parser.parse_args()
    if args.py or args.r or args.oc:
        args.all = False
    try:
        cxx = os.environ['CXX']
    except KeyError:
        os.environ['CXX'] = 'c++'
    main(args)



##############################################################################
## setup.py -- Script to generate swig based interface for python
## Revision: $Id$
##
## Copyright (C) Kshitij Kulshreshtha
##
## This file is part of ADOL-C. This software is provided as open source.
## Any use, reproduction, or distribution of the software constitutes 
## recipient's acceptance of the terms of the accompanying license file.
## 
##############################################################################

from __future__ import print_function
from swigprocess import prepare_flat_header
from numpy.distutils import misc_util as np_dist
from distutils.core import setup, Extension
from distutils.cmd import Command
from distutils.command.build_ext import build_ext
from distutils.command.build import build
from distutils.command.install import install
import os
import subprocess

def compile_dynlib(prefix,colpackdir,boostdir):
    import subprocess
    from multiprocessing import cpu_count
    nproc = cpu_count()
    uppath = lambda _path, n: os.sep.join(_path.split(os.sep)[:-n])
    librootdir = os.path.abspath(uppath(os.getcwd(),2))
    try:
        out = subprocess.check_output(['autoreconf','-fi'],cwd=librootdir,universal_newlines=True)
        print(out)
        out = subprocess.check_output(['./configure',
                                       '--prefix=%s' % prefix,
                                       '--with-colpack=%s' % colpackdir,
                                       '--with-boost=%s' % boostdir ],
                                      cwd=librootdir,
                                      universal_newlines=True)
        print(out)
        out = subprocess.check_output(['make', '-j%d' % nproc, 'install'],cwd=librootdir,
                                      universal_newlines=True)
        print(out)
    except subprocess.CalledProcessError as e:
        print(e.output)
        print('cmd = ', e.cmd)
        raise SystemExit()

class build_adolc(Command):
    user_options = [
        ('lib-prefix=', None, 'prefix to install adolc library'),
        ('colpack-dir=', None, 'directory in which colpack is installed'),
        ('boost-dir=', None, 'directory in which boost is installed'),
        ('only-swig', None, 'only install this swig module, not adolc library') ]
    command_name = 'build_lib'

    def initialize_options(self):
        self.lib_prefix = None
        self.colpack_dir = None
        self.boost_dir = None
        self.only_swig = None

    def finalize_options(self):
        self.set_undefined_options('build',
                                   ('lib_prefix','lib_prefix'),
                                   ('colpack_dir', 'colpack_dir'),
                                   ('boost_dir', 'boost_dir'),
                                   ('only_swig', 'only_swig') )

    def run(self):
        if not self.only_swig:
            compile_dynlib(self.lib_prefix,self.colpack_dir,self.boost_dir)
        else:
            print('skipping build of ADOL-C')


class build_swigadolc(build_ext,object):
    command_name = 'build_ext'
    user_options = build_ext.user_options + [
        ('lib-prefix=', None, 'prefix of installed adolc library') ]
    def initialize_options(self):
        super(build_swigadolc,self).initialize_options()
        self.lib_prefix = None

    def finalize_options(self):
        super(build_swigadolc,self).finalize_options()
        self.set_undefined_options('build',
                                   ('lib_prefix','lib_prefix'))
        prefix = self.lib_prefix
        print('prefix = ', prefix)
        self.include_dirs.append(os.path.join(prefix,'include'))
        self.library_dirs.append(os.path.join(prefix,'lib64'))
        self.rpath.append(os.path.join(prefix,'lib64'))
        prepare_flat_header()


class buildthis(build,object):
    command_name = 'build'
    user_options = build.user_options + [
        ('lib-prefix=', None, 'prefix to install adolc library'),
        ('colpack-dir=', None, 'directory in which colpack is installed'),
        ('boost-dir=', None, 'directory in which boost is installed'),
        ('only-swig', None, 'only install this swig module, not adolc library') ]

    def lib_doesnot_exist(self):
        from distutils.ccompiler import new_compiler
        from distutils.sysconfig import customize_compiler
        compiler = new_compiler()
        customize_compiler(compiler)
        lib_file = compiler.find_library_file([os.path.join(self.lib_prefix,'lib64')],'adolc')
        return lib_file is None

    def initialize_options(self):
        super(buildthis,self).initialize_options()
        self.lib_prefix = None
        self.colpack_dir = None
        self.boost_dir = None
        self.only_swig = None

    #sub_commands = [ ('build_lib', lib_doesnot_exist),
    #                 ('build_ext', None) ]
    sub_commands = [ ('build_lib', None),
                     ('build_ext', None) ]

class installthis(install,object):
    command_name = 'install'
    user_options = install.user_options + [
        ('lib-prefix=', None, 'prefix to install adolc library'),
        ('colpack-dir=', None, 'directory in which colpack is installed'),
        ('boost-dir=', None, 'directory in which boost is installed'),
        ('only-swig', None, 'only install this swig module, not adolc library')]

    def initialize_options(self):
        self.lib_prefix = None
        self.colpack_dir = None
        self.boost_dir = None
        self.only_swig = None
        super(installthis,self).initialize_options()

    def finalize_options(self):
        super(installthis,self).finalize_options()
        if self.lib_prefix is None:
            self.lib_prefix = os.path.join(os.environ['HOME'],'adolc_base')
        if self.colpack_dir is None:
            self.colpack_dir = os.path.join(os.environ['HOME'],'adolc_base')
        if self.boost_dir is None:
            self.boost_dir = '/usr'
        if self.only_swig is None:
            self.only_swig = False
        self.finalized = 1
        buildobj = self.distribution.get_command_obj('build')
        buildobj.set_undefined_options('install',
                                       ('lib_prefix','lib_prefix'),
                                       ('colpack_dir', 'colpack_dir'),
                                       ('boost_dir', 'boost_dir'),
                                       ('only_swig', 'only_swig'))
        
incdirs = np_dist.get_numpy_include_dirs()
#python_ldflags = subprocess.check_output(['python-config','--ldflags'],universal_newlines=True)

adolc_mod = Extension('_adolc',
                      sources=['adolc-python.i'],
                      depends=['adolc_all.hpp'],
                      swig_opts=['-c++', '-dirvtable'],
                      libraries=['adolc'],
                      include_dirs=incdirs,
                      extra_compile_args=['-std=c++11', '-fPIC', '-w'],
                      #extra_link_args=[python_ldflags.rstrip() + ' -Wl,-no-undefined']
)

setup(name='adolc',
      ext_modules=[adolc_mod],
      py_modules=['adolc'],
      version='2.7-trunk',
      cmdclass = { 'build_lib': build_adolc,
                   'build_ext': build_swigadolc,
                   'build': buildthis,
                   'install': installthis
               }
)



from __future__ import print_function
import numpy as np
from adolc import *

def do_taping(tag,x):
    trace_on(tag)
    ax = as_adouble(x)
    for item in iter(ax):
       item.declareIndependent()
    if (ax[0] < 0):
       ay = ax[2]*ax[1]*2.0
    else:
       ay = ax[2]+3.0*ax[1]
    ay.declareDependent()
    trace_off()


if __name__ == '__main__':
    x = [ 1, 2, 3 ]
    disableBranchSwitchWarnings()
    do_taping(1,x)
    g = gradient(1,x)
    print(g)
    x = [ -1, 2, 3]
    try:
       g = gradient(1,x)
    except BranchException: 
       do_taping(1,x)
       g = gradient(1,x)
    print(g)
    



from __future__ import print_function
import adolc
import numpy as np

steps = 100
h = 0.01

def euler_step_act(n,yin,m,yout):
    yout[0] = yin[0] + h*yin[0]
    yout[1] = yin[1] + h*2*yin[1]

class euler_step_edf(adolc.EDFobject):
    def __init__(self,tag):
        super(self.__class__,self).__init__()
        self.tag_ext_fct = tag
    def function(self,arg):
        (n,yin,m,yout) = arg
        yout[0] = yin[0] + h*yin[0]
        yout[1] = yin[1] + h*2*yin[0]
        return 1
    def zos_forward(self,arg):
        (n,yin,m,yout) = arg
        adolc.set_nested_ctx(self.tag_ext_fct,1)
        ny = adolc.zos_forward(self.tag_ext_fct,2,2,0,yin)
        adolc.set_nested_ctx(self.tag_ext_fct,0)
        np.copyto(yout,ny)
        return 1
    def fos_forward(self,arg):
        (n,yin,dyin,m,yout,dyout) = arg
        adolc.set_nested_ctx(self.tag_ext_fct,1)
        ny,ndy = adolc.fos_forward(self.tag_ext_fct,2,2,0,yin,dyin)
        adolc.set_nested_ctx(self.tag_ext_fct,0)
        np.copyto(yout,ny)
        np.copyto(dyout,ndy)
        return 1
    def fov_forward(self,arg):
        (n,yin,p,dyin,m,yout,dyout) = arg
        adolc.set_nested_ctx(self.tag_ext_fct,1)
        ny,ndy = adolc.fov_forward(self.tag_ext_fct,2,2,p,yin,dyin)
        adolc.set_nested_ctx(self.tag_ext_fct,0)
        np.copyto(yout,ny)
        np.copyto(dyout,ndy)
        return 1
    def fos_reverse(self,arg):
        (m,u,n,z,x,y) = arg
        adolc.set_nested_ctx(self.tag_ext_fct,1)
        ny = adolc.zos_forward(self.tag_ext_fct,2,2,1,x)
        nz = adolc.fos_reverse(self.tag_ext_fct,2,2,u)
        np.copyto(z,nz)
        return 1
    def fov_reverse(self,arg):
        (m,q,u,n,z,x,y) = arg
        adolc.set_nested_ctx(self.tag_ext_fct,1)
        ny = adolc.zos_forward(self.tag_ext_fct,2,2,1,x)
        nz = adolc.fov_reverse(self.tag_ext_fct,2,2,q,u)
        np.copyto(z,nz)
        return 1


if __name__ == '__main__':
    tag_full = 1
    tag_part = 2
    tag_ext_fct = 3
    n = 2
    m = 2
    t0 = 0.0
    tf = 1.0
    conp = [ 1.0, 1.0 ]
    adolc.trace_on(tag_full)
    con = adolc.as_adouble(conp)
    for c in con:
        c.declareIndependent()
    y = con
    ynew = adolc.as_adouble([ 0.0, 0.0 ])
    for i in range(steps):
        euler_step_act(n,y,m,ynew)
        y = ynew
    f = y[0] + y[1]
    f.declareDependent()
    adolc.trace_off()
    grad = adolc.gradient(tag_full,conp)
    print(" full taping:\n gradient=(",grad[0],", ",grad[1],")\n\n")

    # now taping external function
    adolc.trace_on(tag_ext_fct)
    con = adolc.as_adouble(conp)
    for c in con:
        c.declareIndependent()
    y = con
    ynew = adolc.as_adouble([ 0.0, 0.0 ])
    euler_step_act(2,y,2,ynew)
    for c in ynew:
        c.declareDependent()
    adolc.trace_off()

    edf = euler_step_edf(tag_ext_fct)    
    # now use the external function
    adolc.trace_on(tag_part)
    con = adolc.advector(conp)
    for c in con:
        c.declareIndependent()
    y = con
    ynew = adolc.advector([0.0,0.0])
    for i in range(steps):
        edf.call(2,y,2,ynew)
        y = ynew
    f = y[0] + y[1]
    f.declareDependent()
    adolc.trace_off()
    gradp = adolc.gradient(tag_part,conp)
    print(" taping with external function facility:\n gradient=(",gradp[0],", ",gradp[1],")\n\n")
